from typing import AsyncGenerator, Dict, Generator, List, Optional, Union

import openai
import requests
from unifyai.exceptions import BadRequestError, UnifyError, status_error_map
from unifyai.utils import (  # noqa:WPS450
    _available_dynamic_modes,
    _validate_api_key,
    _validate_endpoint,
)


class Unify:
    """Class for interacting with the Unify API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        endpoint: Optional[str] = None,
        model: Optional[str] = None,
        provider: Optional[str] = None,
    ) -> None:  # noqa: DAR101, DAR401
        """Initialize the Unify client.

        Args:
            api_key (str, optional): API key for accessing the Unify API.
                If None, it attempts to retrieve the API key from the
                environment variable UNIFY_KEY.
                Defaults to None.

            endpoint (str, optional): Endpoint name in OpenAI API format:
                <uploaded_by>/<model_name>@<provider_name>
                Defaults to None.

            model (str, optional): Name of the model. If None,
            endpoint must be provided.

            provider (str, optional): Name of the provider. If None,
            endpoint must be provided.
        Raises:
            UnifyError: If the API key is missing.
        """
        self._api_key = _validate_api_key(api_key)
        self._endpoint, self._model, self._provider = _validate_endpoint(  # noqa:WPS414
            endpoint,
            model,
            provider,
        )
        try:
            self.client = openai.OpenAI(
                base_url="https://api.unify.ai/v0/",
                api_key=self._api_key,
            )
        except openai.OpenAIError as e:
            raise UnifyError(f"Failed to initialize Unify client: {str(e)}")

    @property
    def model(self) -> str:
        """
        Get the model name.  # noqa: DAR201.

        Returns:
            str: The model name.
        """
        return self._model

    def set_model(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The model name.
        """
        self._model = value
        if self._provider:
            self._endpoint = "@".join([value, self._provider])
        else:
            mode = self._endpoint.split("@")[1]
            self._endpoint = "@".join([value, mode])

    @property
    def provider(self) -> Optional[str]:
        """
        Get the provider name.  # noqa :DAR201.

        Returns:
            str: The provider name.
        """
        return self._provider

    def set_provider(self, value: str) -> None:
        """
        Set the provider name.  # noqa: DAR101.

        Args:
            value (str): The provider name.
        """
        self._provider = value
        self._endpoint = "@".join([self._model, value])

    @property
    def endpoint(self) -> str:
        """
        Get the endpoint name.  # noqa: DAR201.

        Returns:
            str: The endpoint name.
        """
        return self._endpoint

    def set_endpoint(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The endpoint name.
        """
        self._endpoint = value
        self._model, self._provider = value.split("@")  # noqa: WPS414
        if self._provider in _available_dynamic_modes:
            self._provider = None

    def generate(  # noqa: WPS234, WPS211
        self,
        messages: Optional[List[Dict[str, str]]] = None,
        user_prompt: Optional[str] = None,
        system_prompt: Optional[str] = None,
        stream: bool = False,
    ) -> Union[Generator[str, None, None], str]:  # noqa: DAR101, DAR201, DAR401
        """Generate content using the Unify API.

        Args:
            messages (List[Dict[str, str]]): A list of dictionaries containing the
            conversation history. If provided, user_prompt must be None.

            user_prompt (Optional[str]): A string containing the user prompt.
            If provided, messages must be None.

            system_prompt (Optional[str]): An optional string containing the
            system prompt.

            stream (bool): If True, generates content as a stream.
            If False, generates content as a single response.
            Defaults to False.

        Returns:
            Union[Generator[str, None, None], str]: If stream is True,
             returns a generator yielding chunks of content.
             If stream is False, returns a single string response.

        Raises:
            UnifyError: If an error occurs during content generation.
        """
        contents = []
        if system_prompt:
            contents.append({"role": "system", "content": system_prompt})

        if user_prompt:
            contents.append({"role": "user", "content": user_prompt})
        elif messages:
            contents.extend(messages)
        else:
            raise UnifyError("You must provider either the user_prompt or messages!")

        if stream:
            return self._generate_stream(contents, self._endpoint)
        return self._generate_non_stream(contents, self._endpoint)

    def get_credit_balance(self) -> Optional[int]:
        # noqa: DAR201, DAR401
        """
        Get the remaining credits left on your account.

        Returns:
            int or None: The remaining credits on the account
            if successful, otherwise None.
        Raises:
            BadRequestError: If there was an HTTP error.
            ValueError: If there was an error parsing the JSON response.
        """
        url = "https://api.unify.ai/v0/get_credits"
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        try:
            response = requests.get(url, headers=headers, timeout=10)
            response.raise_for_status()
            return response.json()["credits"]
        except requests.RequestException as e:
            raise BadRequestError("There was an error with the request.") from e
        except (KeyError, ValueError) as e:
            raise ValueError("Error parsing JSON response.") from e

    def _generate_stream(
        self,
        messages: List[Dict[str, str]],
        endpoint: str,
    ) -> Generator[str, None, None]:
        try:
            chat_completion = self.client.chat.completions.create(
                model=endpoint,
                messages=messages,  # type: ignore[arg-type]
                stream=True,
            )
            for chunk in chat_completion:
                content = chunk.choices[0].delta.content  # type: ignore[union-attr]
                self._provider = chunk.model.split("@")[-1]  # type: ignore[union-attr]
                if content is not None:
                    yield content
        except openai.APIStatusError as e:
            raise status_error_map[e.status_code](e.message) from None

    def _generate_non_stream(
        self,
        messages: List[Dict[str, str]],
        endpoint: str,
    ) -> str:
        try:
            chat_completion = self.client.chat.completions.create(
                model=endpoint,
                messages=messages,  # type: ignore[arg-type]
                stream=False,
            )
            self._provider = chat_completion.model.split(  # type: ignore[union-attr]
                "@",
            )[-1]

            return chat_completion.choices[0].message.content.strip(" ")  # type: ignore # noqa: E501, WPS219
        except openai.APIStatusError as e:
            raise status_error_map[e.status_code](e.message) from None


class AsyncUnify:
    """Class for interacting asynchronously with the Unify API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        endpoint: Optional[str] = None,
        model: Optional[str] = None,
        provider: Optional[str] = None,
    ) -> None:  # noqa:DAR101, DAR401
        """Initialize the AsyncUnify client.

        Args:
            api_key (str, optional): API key for accessing the Unify API.
            If None, it attempts to retrieve the API key from
            the environment variable UNIFY_KEY.
            Defaults to None.

            endpoint (str, optional): Endpoint name in OpenAI API format:
            <uploaded_by>/<model_name>@<provider_name>
            Defaults to None.

            model (str, optional): Name of the model. If None,
            endpoint must be provided.

            provider (str, optional): Name of the provider. If None,
            endpoint must be provided.

        Raises:
            UnifyError: If the API key is missing.
        """
        self._api_key = _validate_api_key(api_key)
        self._endpoint, self._model, self._provider = (  # noqa: WPS414
            _validate_endpoint(
                endpoint,
                model,
                provider,
            )
        )
        try:
            self.client = openai.AsyncOpenAI(
                base_url="https://api.unify.ai/v0/",
                api_key=self._api_key,
            )
        except openai.APIStatusError as e:
            raise UnifyError(f"Failed to initialize Unify client: {str(e)}")

    @property
    def model(self) -> str:
        """
        Get the model name.  # noqa: DAR201.

        Returns:
            str: The model name.
        """
        return self._model

    def set_model(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The model name.
        """
        self._model = value
        if self._provider:
            self._endpoint = "@".join([value, self._provider])
        else:
            mode = self._endpoint.split("@")[1]
            self._endpoint = "@".join([value, mode])

    @property
    def provider(self) -> Optional[str]:
        """
        Get the provider name.  # noqa :DAR201.

        Returns:
            str: The provider name.
        """
        return self._provider

    def set_provider(self, value: str) -> None:
        """
        Set the provider name.  # noqa: DAR101.

        Args:
            value (str): The provider name.
        """
        self._provider = value
        self._endpoint = "@".join([self._model, value])

    @property
    def endpoint(self) -> str:
        """
        Get the endpoint name.  # noqa: DAR201.

        Returns:
            str: The endpoint name.
        """
        return self._endpoint

    def set_endpoint(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The endpoint name.
        """
        self._endpoint = value
        self._model, self._provider = value.split("@")  # noqa: WPS414
        if self._provider in _available_dynamic_modes:
            self._provider = None

    def get_credit_balance(self) -> Optional[int]:
        # noqa: DAR201, DAR401
        """
        Get the remaining credits left on your account.

        Returns:
            int or None: The remaining credits on the account
            if successful, otherwise None.

        Raises:
            BadRequestError: If there was an HTTP error.
            ValueError: If there was an error parsing the JSON response.
        """
        url = "https://api.unify.ai/v0/get_credits"
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        try:
            response = requests.get(url, headers=headers, timeout=10)
            response.raise_for_status()
            return response.json()["credits"]
        except requests.RequestException as e:
            raise BadRequestError("There was an error with the request.") from e
        except (KeyError, ValueError) as e:
            raise ValueError("Error parsing JSON response.") from e

    async def generate(  # noqa: WPS234, WPS211
        self,
        messages: Optional[List[Dict[str, str]]] = None,
        user_prompt: Optional[str] = None,
        system_prompt: Optional[str] = None,
        stream: bool = False,
    ) -> Union[AsyncGenerator[str, None], str]:  # noqa: DAR101, DAR201, DAR401
        """Generate content asynchronously using the Unify API.

        Args:
            messages (List[Dict[str, str]]): A list of dictionaries containing the
            conversation history. If provided, user_prompt must be None.

            user_prompt (Optional[str]): A string containing the user prompt.
            If provided, messages must be None.

            system_prompt (Optional[str]): An optional string containing the
            system prompt.

            stream (bool): If True, generates content as a stream.
            If False, generates content as a single response.
            Defaults to False.

        Returns:
            Union[AsyncGenerator[str, None], List[str]]: If stream is True,
            returns an asynchronous generator yielding chunks of content.
            If stream is False, returns a list of string responses.

        Raises:
            UnifyError: If an error occurs during content generation.
        """
        contents = []
        if system_prompt:
            contents.append({"role": "system", "content": system_prompt})

        if user_prompt:
            contents.append({"role": "user", "content": user_prompt})
        elif messages:
            contents.extend(messages)
        else:
            raise UnifyError("You must provide either the user_prompt or messages!")

        if stream:
            return self._generate_stream(contents, self._endpoint)
        return await self._generate_non_stream(contents, self._endpoint)

    async def _generate_stream(
        self,
        messages: List[Dict[str, str]],
        endpoint: str,
    ) -> AsyncGenerator[str, None]:
        try:
            async_stream = await self.client.chat.completions.create(
                model=endpoint,
                messages=messages,  # type: ignore[arg-type]
                stream=True,
            )
            async for chunk in async_stream:  # type: ignore[union-attr]
                self._provider = chunk.model.split("@")[-1]
                yield chunk.choices[0].delta.content or ""
        except openai.APIStatusError as e:
            raise status_error_map[e.status_code](e.message) from None

    async def _generate_non_stream(
        self,
        messages: List[Dict[str, str]],
        endpoint: str,
    ) -> str:
        try:
            async_response = await self.client.chat.completions.create(
                model=endpoint,
                messages=messages,  # type: ignore[arg-type]
                stream=False,
            )
            self._provider = async_response.model.split("@")[-1]  # type: ignore
            return async_response.choices[0].message.content.strip(" ")  # type: ignore # noqa: E501, WPS219
        except openai.APIStatusError as e:
            raise status_error_map[e.status_code](e.message) from None



import sys

from unifyai.clients import Unify


class ChatBot:  # noqa: WPS338
    """Agent class represents an LLM chat agent."""

    def __init__(self, endpoint: str) -> None:
        """
        Initializes the ChatBot object.

        Args:
            endpoint (str): The endpoint for the chatbot.
        """
        self._message_history = []
        self._endpoint = endpoint
        self._paused = False
        self._client = Unify(
            endpoint=self._endpoint,
        )

    def _get_credits(self):
        """
        Retrieves the current credit balance from associated with the UNIFY account.

        Returns:
            float: Current credit balance.
        """
        return self._client.get_credit_balance()

    def _process_input(self, inp: str, show_credits: bool, show_provider: bool):
        """
        Processes the user input to generate AI response.

        Args:
            inp (str): User input message.
            show_credits (bool): Whether to show credit consumption.

        Yields:
            str: Generated AI response chunks.
        """
        self._update_message_history(inp)
        initial_credit_balance = self._get_credits()
        stream = self._client.generate(
            messages=self._message_history,
            stream=True,
        )
        words = ""
        for chunk in stream:
            words += chunk
            yield chunk

        self._message_history.append(
            {
                "role": "assistant",
                "content": words,
            },
        )
        final_credit_balance = self._get_credits()
        if show_credits:
            sys.stdout.write(
                "\n(spent {:.6f} credits)".format(
                    initial_credit_balance - final_credit_balance,
                ),
            )
        if show_provider:
            sys.stdout.write("\n(provider: {})".format(self._client.provider))

    def _update_message_history(self, inp):
        """
        Updates message history with user input.

        Args:
            inp (str): User input message.
        """
        self._message_history.append(
            {
                "role": "user",
                "content": inp,
            },
        )

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        self._endpoint = value
        self._client.set_endpoint(self._endpoint)

    def clear_chat_history(self):
        """Clears the chat history."""
        self._message_history.clear()

    def run(self, show_credits: bool = False, show_provider: bool = False):
        """
        Starts the chat interaction loop.

        Args:
            show_credits (bool, optional): Whether to show credit consumption. Defaults to False.
        """
        if not self._paused:
            sys.stdout.write(
                "Let's have a chat. (Enter `pause` to pause and `quit` to exit)\n",
            )
            self.clear_chat_history()
        else:
            sys.stdout.write(
                "Welcome back! (Remember, enter `pause` to pause and `quit` to exit)\n",
            )
        self._paused = False
        while True:
            sys.stdout.write("> ")
            inp = input()
            if inp == "quit":
                self.clear_chat_history()
                break
            elif inp == "pause":
                self._paused = True
                break
            for word in self._process_input(inp, show_credits, show_provider):
                sys.stdout.write(word)
                sys.stdout.flush()
            sys.stdout.write("\n")



import os
from typing import Optional, Tuple

from unifyai.exceptions import UnifyError

_available_dynamic_modes = [
    "lowest-input-cost",
    "input-cost",
    "lowest-output-cost",
    "output-cost",
    "lowest-itl",
    "itl",
    "lowest-ttft",
    "ttft",
    "highest-tks-per-sec",
    "tks-per-sec",
]


def _validate_api_key(api_key: Optional[str]) -> str:
    if api_key is None:
        api_key = os.environ.get("UNIFY_KEY")
    if api_key is None:
        raise KeyError(
            "UNIFY_KEY is missing. Please make sure it is set correctly!",
        )
    return api_key


def _validate_endpoint_name(value: str) -> Tuple[str, str]:
    error_message = "endpoint string must use OpenAI API format: <uploaded_by>/<model_name>@<provider_name>"  # noqa: E501

    if not isinstance(value, str):
        raise UnifyError(error_message)

    try:
        model_name, provider_name = value.split("/")[-1].split("@")
    except ValueError:
        raise UnifyError(error_message)

    if not model_name or not provider_name:
        raise UnifyError(error_message)
    return (model_name, provider_name)


def _validate_endpoint(  # noqa: WPS231
    endpoint: Optional[str] = None,
    model: Optional[str] = None,
    provider: Optional[str] = None,
) -> Tuple[str, str, Optional[str]]:
    error_message = (
        "You must either provide an endpoint or the model and provider names!"
    )
    if endpoint:
        if model or provider:
            raise UnifyError(error_message)
        model, provider = _validate_endpoint_name(endpoint)  # noqa: WPS414
    else:
        if not model or not provider:
            raise UnifyError(error_message)
        endpoint = "@".join([model, provider])

    if provider in _available_dynamic_modes:
        provider = None
    return endpoint, model, provider



import os
import unittest
from types import AsyncGeneratorType, GeneratorType
from unittest.mock import MagicMock, patch

from unifyai.clients import AsyncUnify, Unify
from unifyai.exceptions import AuthenticationError, UnifyError


class TestUnify(unittest.TestCase):
    def setUp(self) -> None:
        # Set up a valid API key for testing
        self.valid_api_key = os.environ.get("UNIFY_KEY")

    def test_invalid_api_key_raises_authentication_error(self) -> None:
        # Instantiate Unify with an invalid API key
        with self.assertRaises(AuthenticationError):
            unify = Unify(
                api_key="invalid_api_key",
                endpoint="llama-2-7b-chat@anyscale",
            )
            unify.generate(user_prompt="hello")

    @patch("os.environ.get", return_value=None)
    def test_missing_api_key_raises_key_error(self, mock_get: MagicMock) -> None:
        # Initializing Unify without providing API key should raise KeyError
        with self.assertRaises(KeyError):
            Unify(endpoint="llama-2-7b-chat@anyscale")

    def test_incorrect_model_name_raises_internal_server_error(self) -> None:
        # Provide incorrect model name
        with self.assertRaises(UnifyError):
            Unify(api_key=self.valid_api_key, model="llama-chat")

    def test_generate_returns_string_when_stream_false(self) -> None:
        # Instantiate Unify with a valid API key
        unify = Unify(api_key=self.valid_api_key, endpoint="llama-2-7b-chat@anyscale")
        # Call generate with stream=False
        result = unify.generate(user_prompt="hello", stream=False)
        # Assert that the result is a string
        self.assertIsInstance(result, str)

    def test_generate_returns_generator_when_stream_true(self) -> None:
        # Instantiate Unify with a valid API key
        unify = Unify(api_key=self.valid_api_key, endpoint="llama-2-7b-chat@anyscale")
        # Call generate with stream=True
        result = unify.generate(user_prompt="hello", stream=True)
        # Assert that the result is a generator
        self.assertIsInstance(result, GeneratorType)


class TestAsyncUnify(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        # Set up a valid API key for testing
        self.valid_api_key = os.environ.get("UNIFY_KEY")

    async def test_invalid_api_key_raises_authentication_error(self) -> None:
        # Instantiate AsyncUnify with an invalid API key
        with self.assertRaises(AuthenticationError):
            async_unify = AsyncUnify(
                api_key="invalid_api_key",
                endpoint="llama-2-7b-chat@anyscale",
            )
            await async_unify.generate(user_prompt="hello")

    @patch("os.environ.get", return_value=None)
    async def test_missing_api_key_raises_key_error(self, mock_get: MagicMock) -> None:
        # Initializing AsyncUnify without providing
        # API key should raise KeyError
        with self.assertRaises(KeyError):
            async_unify = AsyncUnify()
            await async_unify.generate(user_prompt="hello")

    async def test_incorrect_model_name_raises_internal_server_error(self) -> None:
        # Provide incorrect model name
        with self.assertRaises(UnifyError):
            AsyncUnify(api_key=self.valid_api_key, model="llama-chat")

    async def test_generate_returns_string_when_stream_false(self) -> None:
        # Instantiate AsyncUnify with a valid API key
        async_unify = AsyncUnify(
            api_key=self.valid_api_key,
            endpoint="llama-2-7b-chat@anyscale",
        )
        # Call generate with stream=False
        result = await async_unify.generate(user_prompt="hello", stream=False)
        # Assert that the result is a string
        self.assertIsInstance(result, str)

    async def test_generate_returns_generator_when_stream_true(self) -> None:
        # Instantiate AsyncUnify with a valid API key
        async_unify = AsyncUnify(
            api_key=self.valid_api_key,
            endpoint="llama-2-7b-chat@anyscale",
        )
        # Call generate with stream=True
        result = await async_unify.generate(user_prompt="hello", stream=True)
        # Assert that the result is a generator
        self.assertIsInstance(result, AsyncGeneratorType)


if __name__ == "__main__":
    unittest.main()



class UnifyError(Exception):
    """Base class for all custom exceptions in the Unify application."""


class BadRequestError(UnifyError):
    """Exception raised for HTTP 400 Bad Request errors."""


class AuthenticationError(UnifyError):
    """Exception raised for HTTP 401 Unauthorized errors."""


class PermissionDeniedError(UnifyError):
    """Exception raised for HTTP 403 Forbidden errors."""


class NotFoundError(UnifyError):
    """Exception raised for HTTP 404 Not Found errors."""


class ConflictError(UnifyError):
    """Exception raised for HTTP 409 Conflict errors."""


class UnprocessableEntityError(UnifyError):
    """Exception raised for HTTP 422 Unprocessable Entity errors."""


class RateLimitError(UnifyError):
    """Exception raised for HTTP 429 Too Many Requests errors."""


class InternalServerError(UnifyError):
    """Exception raised for HTTP 500 Internal Server Error errors."""


status_error_map = {
    400: BadRequestError,
    401: AuthenticationError,
    403: PermissionDeniedError,
    404: NotFoundError,
    409: ConflictError,
    422: UnprocessableEntityError,
    429: RateLimitError,
    500: InternalServerError,
}
